package com.mh.user.service.impl;

import com.alibaba.fastjson2.JSON;
import com.alibaba.fastjson2.JSONArray;
import com.alibaba.fastjson2.JSONObject;
import com.alibaba.fastjson2.JSONWriter;
import com.github.benmanes.caffeine.cache.Cache;
import com.mh.algorithm.bpnn.BPModel;
import com.mh.algorithm.bpnn.BPNeuralNetworkFactory;
import com.mh.algorithm.bpnn.BPParameter;
import com.mh.algorithm.matrix.Matrix;
import com.mh.algorithm.utils.CsvInfo;
import com.mh.algorithm.utils.SerializationUtil;
import com.mh.common.utils.StringUtils;
import com.mh.user.dto.EnergyPreDTO;
import com.mh.user.dto.EnergyPreEchartDataDTO;
import com.mh.user.dto.EnergyPreTopDataDTO;
import com.mh.user.entity.HistoryDataPre;
import com.mh.user.entity.SysParamEntity;
import com.mh.user.job.GetWeatherInfoJob;
import com.mh.user.mapper.HistoryDataPreMapper;
import com.mh.user.service.HistoryDataPreService;
import com.mh.user.service.SysParamService;
import com.mh.user.utils.DateUtil;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

/**
 * @author LJF
 * @version 1.0
 * @project CHWS
 * @description 预测历史数据服务实现类
 * @date 2024-05-09 10:03:24
 */
@Service
@Transactional(rollbackFor = Exception.class)
public class HistoryDataPreServiceImpl implements HistoryDataPreService {

    @Resource
    private HistoryDataPreMapper historyDataPreMapper;

    @Resource
    @Qualifier("caffeineCache")
    private Cache caffeineCache;

    @Resource
    private SysParamService sysParamService;

    @Resource
    private GetWeatherInfoJob getWeatherInfoJob;


    public static String[] convert(HistoryDataPre dataPre) {
        // 假设HistoryDataPre有字段如field1, field2, field3等，根据需要进行转换
        return new String[]{
                String.valueOf(dataPre.getEnvMinTemp()),
                String.valueOf(dataPre.getEnvMaxTemp()),
                String.valueOf(dataPre.getPeopleNum()),
                String.valueOf(dataPre.getWaterValue()),
                String.valueOf(dataPre.getElectValue()),
                String.valueOf(dataPre.getWaterLevel())
        };
    }

    @Override
    public void startTrainData(String buildingId) throws Exception {
        List<HistoryDataPre> trainData = historyDataPreMapper.getTrainData(buildingId);
        if (trainData == null || trainData.size() == 0) {
            return;
        }
        List<String[]> historyDataPreList = new ArrayList<>();
        for (HistoryDataPre dataPre : trainData) {
            historyDataPreList.add(convert(dataPre));
        }
        // 创建训练集矩阵
        CsvInfo csvInfo = new CsvInfo();
        csvInfo.setCsvFileList(new ArrayList<>(historyDataPreList));
        Matrix trainSet = csvInfo.toMatrix();

        // 创建BPNN工厂对象
        BPNeuralNetworkFactory factory = new BPNeuralNetworkFactory();

        // 创建BP参数对象
        BPParameter bpParameter = new BPParameter();
        bpParameter.setInputLayerNeuronCount(3);
        bpParameter.setHiddenLayerNeuronCount(3);
        bpParameter.setOutputLayerNeuronCount(3);
        bpParameter.setPrecision(0.01);
        bpParameter.setMaxTimes(50000);

        // 训练BP神经网络
        BPModel bpModel = factory.trainBP(bpParameter, trainSet);

        // 将BPModel序列化到本地
        SerializationUtil.serialize(bpModel, buildingId + "_pre_data");
    }

    @Override
    public void startPredictData(String buildingId, String curDate) throws Exception {
        // 判断是否存在天气温度数据以及现在的用水量用电量等
//        int isPre = historyDataPreMapper.selectIsPre(buildingId, curDate);
//        if (isPre > 0) {
//            return;
//        }
        // 获取当前天气数据
        SysParamEntity sysParam = sysParamService.selectSysParam();
        Object weather = caffeineCache.getIfPresent(sysParam.getProArea());
        if (weather == null) {
            getWeatherInfoJob.getWeatherInfo();
            weather = caffeineCache.getIfPresent(sysParam.getProArea());
        }
        String weatherStr = (String) weather;
        JSONObject jsonObject = JSON.parseObject(weatherStr);
        if (null == jsonObject) {
            return;
        }
        String envMinTemp = "16.50";
        String envMaxTemp = "26.00";
        JSONArray jsonArray = jsonObject.getJSONArray("forecasts").getJSONObject(0).getJSONArray("casts");
        for (int i = 0; i < jsonArray.size(); i++) {
            JSONObject jsonObject1 = jsonArray.getJSONObject(i);
            if (jsonObject1.getString("date").equals(curDate)) {
                envMinTemp = jsonObject1.getString("nighttemp");
                envMaxTemp = jsonObject1.getString("daytemp");
                break;
            }
        }
        // 获取当前用水量和用电量以及实际平均水位
        HistoryDataPre curHistoryData = historyDataPreMapper.selectCurData(buildingId, curDate);
        // 插入数据
        curHistoryData.setEnvMaxTemp(new BigDecimal(envMaxTemp));
        curHistoryData.setEnvMinTemp(new BigDecimal(envMinTemp));
        HistoryDataPre historyDataPre = historyDataPreMapper.selectOneData(buildingId, curDate);
        if (historyDataPre == null) {
            historyDataPreMapper.insertData(curHistoryData);
        }
        // 开始预测
        HistoryDataPre historyDataPre1 = historyDataPreMapper.selectOneData(buildingId, curDate);
        String[] preData = new String[]{
                historyDataPre1.getEnvMinTemp().toString(),
                historyDataPre1.getEnvMaxTemp().toString(),
                historyDataPre1.getPeopleNum().toString()
        };

        CsvInfo csvInfo = new CsvInfo();
        ArrayList<String[]> list = new ArrayList<>();
        list.add(preData);
        csvInfo.setCsvFileList(list);
        Matrix data = csvInfo.toMatrix();
        // 将BPModel反序列化
        BPModel bpModel1 = (BPModel) SerializationUtil.deSerialization(buildingId + "_pre_data");
        // 创建工厂
        BPNeuralNetworkFactory factory = new BPNeuralNetworkFactory();
        Matrix result = factory.computeBP(bpModel1, data);
        // 得出预测数据
        HistoryDataPre preHistoryData = new HistoryDataPre();
        preHistoryData.setId(historyDataPre1.getId());
        preHistoryData.setBuildingId(buildingId);
        for (int i = 0; i < result.getMatrixRowCount(); i++) {
            String[] record = new String[result.getMatrixColCount()];
            for (int j = 0; j < result.getMatrixColCount(); j++) {
                record[j] = String.valueOf(result.getValOfIdx(i, j));
            }
            // 拼接预测值
            preHistoryData.setWaterValuePre(evaluateAndReturnBigDecimal(record[0]));
            preHistoryData.setElectValuePre(evaluateAndReturnBigDecimal(record[1]));
            preHistoryData.setWaterLevelPre(evaluateAndReturnBigDecimal(record[2]).compareTo(BigDecimal.valueOf(100)) > 0 ? BigDecimal.valueOf(100) : evaluateAndReturnBigDecimal(record[2]));
        }
        preHistoryData.setWaterValue(curHistoryData.getWaterValue());
        preHistoryData.setElectValue(curHistoryData.getElectValue());
        preHistoryData.setWaterLevel(curHistoryData.getWaterLevel());
        // 更新预测值
        historyDataPreMapper.updateById(preHistoryData);
    }

    /**
     * 判断输入的字符串转换的BigDecimal是否小于或等于0，返回BigDecimal.ZERO。
     * 如果小于或等于0，返回输入的BigDecimal；如果大于0，然后返回相应的BigDecimal值。
     *
     * @param recordValue 记录中的值，用于转换为BigDecimal进行比较
     * @return 经过判断后的BigDecimal值
     */
    public static BigDecimal evaluateAndReturnBigDecimal(String recordValue) {
        BigDecimal value = new BigDecimal(recordValue);
        if (value.compareTo(BigDecimal.ZERO) >= 0) {
            return value.setScale(2, RoundingMode.HALF_UP);
        } else {
            return BigDecimal.ZERO;
        }
    }

    @Override
    public List<HistoryDataPre> getRecentData(String buildingId, String curDate) {
        return historyDataPreMapper.getRecentData(buildingId, curDate);
    }

    @Override
    public List<HashMap<String, Object>> getEnergyPre(String buildingId, String beginDate, String endDate, String type) {
        if (StringUtils.isBlank(beginDate) || StringUtils.isBlank(endDate)) {
            DateTimeFormatter formatter = DateTimeFormatter.ofPattern("yyyy-MM-dd");
            // 获取当前日期
            LocalDate now = LocalDate.now();
            // 向前推30天
            LocalDate startDate = now.minusDays(30);
            beginDate = startDate.format(formatter);
            // 结束日期是当前日期
            endDate = now.format(formatter);
        }
        if (StringUtils.isBlank(buildingId) || StringUtils.isBlank(type)) {
            return null;
        }
        List<EnergyPreEchartDataDTO> energyPre = historyDataPreMapper.getEnergyPre(buildingId, beginDate, endDate, type);
        if (energyPre.size() == 0) {
            return null;
        }
        String[] curDate = energyPre.stream()
                .map(EnergyPreEchartDataDTO::getCurDate)
                .toArray(String[]::new); // 使用stream和map转换每个对象的特定字段为JSON字符串，然后转换成数组
        String[] curData = energyPre.stream()
                .map(EnergyPreEchartDataDTO::getCurData)
                .toArray(String[]::new); // 使用stream和map转换每个对象的特定字段为JSON字符串，然后转换成数组
        String[] preData = energyPre.stream()
                .map(EnergyPreEchartDataDTO::getPreData)
                .toArray(String[]::new); // 使用stream和map转换每个对象的特定字段为JSON字符串，然后转换成数组
        String[] errorData = energyPre.stream()
                .map(EnergyPreEchartDataDTO::getErrorData)
                .toArray(String[]::new); // 使用stream和map转换每个对象的特定字段为JSON字符串，然后转换成数组
        List<HashMap<String, Object>> resultList = new ArrayList<>();
        HashMap<String, Object> resultHashMap = new HashMap<>();
        resultHashMap.put("curDate", curDate);
        resultHashMap.put("curData", curData);
        resultHashMap.put("preData", preData);
        resultHashMap.put("errorData", errorData);
        resultList.add(resultHashMap);
        return resultList;
    }

    @Override
    public List<EnergyPreTopDataDTO> getTopData(String buildingId, String type) {
        // 获取顶部数据（昨日，昨日预测，今日预测，昨日偏差值）
        return historyDataPreMapper.getTopData(buildingId, type);
    }
}
